/* $Id: generic.c,v 1.7 1999/06/21 23:44:57 marcus Exp $
******************************************************************************

   KGI - S3 generic acceleration driver

   Copyright (C) 1995-1997 Andreas Beck		[becka@ggi-project.org]
   Copyright (C) 1995-1997 Steffen Seeger	[seeger@ggi-project.org]

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************************
*/
 
#error \
This driver is outdated. If you are using a S3-Vision 96x select \
accel -> S3 -> 86c964.c, for every other chipset use \
accel -> generic -> generic

#include <kgi/maintainers.h>
#define	MAINTAINER	Andreas_Beck
#define	DRIVER_REV	"$Revision: 1.7 $"

/*\
|*| KGI includes
\*/

#include <kgi/module.h>
#include <kgi/kgi_commands.h>
#include <kgi/graphic.h>

#include "suggest.inc"


#define	request_io_region	request_region
#define	check_io_region		check_region


#define	VISION964
#define VISION968
#include <S3/Vision.h>

#include <S3/accel.h>

#define NUMPINGPONGPAGES 2

void *pingpong;
enum pingpongstatus 
{ 
	WAITING, RECEIVING, EXECUTING 
};

enum pingpongstatus ppstat[NUMPINGPONGPAGES];

/* Helper functions */
int s3_wait_fifo(int spc_need, int maysched)
{	
	int x = 1000000000;
	while((inb(ECR_GP_STATUS) & (0x100 >> spc_need)) && --x) 
		if (maysched) schedule();
	return(!x);
}

static int s3_wait_idle(int maysched)
{ 	
	int x = 1000000000;
	while((inw(ECR_GP_STATUS) & EC06_GE_BUSY) && --x) 
		if (maysched) schedule();
	return(!x);
}

/* Accelerated drawing functions follow */
static int s3_drawbox(struct kgi_graphic *graphic, void *iobuf, int maysched)
{	
	struct kgi_box *boxp = (struct kgi_box *)iobuf;

	if (graphic->dev.mode.graphtype == KGIGT_4BIT || 
	    graphic->dev.mode.graphtype == KGIGT_TEXT16)
	{
		
		return -E(ACCEL, NOSUP_MODE_MMAP);
	}

	if (boxp->w && boxp->h) {	/* 0 width is not OK! */

		if (s3_wait_fifo(3, 1))
		{
			
			return -E(ACCEL,NOSUP_NOW_MMAP);
		}
		
		outw(0x27, ECR_FOREGR_MIX);
		outl(graphic->gc->fg_color, ECR_FOREGROUND);
		outw(MF_PIX_CONTROL, ECR_MULTIFUNC);

		if (s3_wait_fifo(5, maysched))
		{
			
			return -E(ACCEL,NOSUP_NOW_MMAP);
		}
		
		outw(boxp->x, ECR_CUR_X);
		outw(boxp->y, ECR_CUR_Y);
		outw(boxp->w - 1, ECR_MAJAXIS_CNT);
		outw((boxp->h - 1) | MF_MINAXIS_CNT, ECR_MULTIFUNC);
		outw(EC07_FILLRECT | EC07_LITTLE_ENDIAN |  EC07_32BIT_IO |
		     EC07_DIR(5) | EC07_DRAWPIXEL | EC07_ACROSS_PLANE |
	             0x0001, ECR_COMMAND);
		s3_wait_idle(maysched);
	}

	return(0);
}

static int s3_copybox(struct kgi_graphic *graphic,void *iobuf)
{	
	struct kgi_cbox *boxp = (struct kgi_cbox *)iobuf;
	int dir = 0;

	if (graphic->dev.mode.graphtype == KGIGT_4BIT || 
	    graphic->dev.mode.graphtype == KGIGT_TEXT16)
		return -E(ACCEL,NOSUP_MODE_MMAP);

	if (boxp->w&&boxp->h) {		/* 0 width is not OK! */

		if (s3_wait_fifo(2, 1)) 
		{
			
			return -E(ACCEL,NOSUP_NOW_MMAP);
		}

		outw(0x67, ECR_FOREGR_MIX);
		outw(MF_PIX_CONTROL, ECR_MULTIFUNC);

		if (boxp->x > boxp->x2) 
		{
			
			dir |= EC07_DIR(1);
		}
		else 
		{
			
			boxp->x += boxp->w - 1;
			boxp->x2 += boxp->w - 1;
		}

		if (boxp->y > boxp->y2) 
		{
			
			dir |= EC07_DIR(4);
		}
		else 
		{ 
			boxp->y += boxp->h - 1;
			boxp->y2 += boxp->h - 1;
		}

		if (s3_wait_fifo(7, 1)) 
		{
			
			return -E(ACCEL,NOSUP_NOW_MMAP);
		}
		
		outw(boxp->x, ECR_CUR_X);
		outw(boxp->y, ECR_CUR_Y);
		outw(boxp->x2, ECR_DIAG_STEP);
		outw(boxp->y2, ECR_AXIAL_STEP);
		outw(boxp->w - 1, ECR_MAJAXIS_CNT);
		outw((boxp->h - 1) | MF_MINAXIS_CNT,ECR_MULTIFUNC);
		outw(EC07_BITBLT | EC07_LITTLE_ENDIAN | EC07_32BIT_IO |
		     dir | EC07_DRAWPIXEL | EC07_ACROSS_PLANE |
	             0x0001, ECR_COMMAND);
		s3_wait_idle(1);
	}
	
	return(0);
}

static int s3_drawline(struct kgi_graphic *graphic, void *iobuf)
{ 
	int dx, dy, min, max;
	struct kgi_line *linp = (struct kgi_line *)iobuf;

	if (graphic->dev.mode.graphtype == KGIGT_4BIT || 
	    graphic->dev.mode.graphtype == KGIGT_TEXT16)
	{
		
		return -E(ACCEL,NOSUP_MODE_MMAP);
	}

	dx = linp->x2 - linp->x;
	dy = linp->y2 - linp->y;

	if (abs(dx) > abs(dy))
	{
		
		max = abs(dx);
		min = abs(dy);
	}
	else
	{
		
		max = abs(dy);
		min = abs(dx);
	}

	if (max == 0)
	{
		
		return(0);	/* FIXME! */
	}

	if (s3_wait_fifo(3, 1))
	{ 
		
		return -E(ACCEL,NOSUP_NOW_MMAP);
	}

	outw(0x27, ECR_FOREGR_MIX);
	outl(graphic->gc->fg_color, ECR_FOREGROUND);
	outw(MF_PIX_CONTROL, ECR_MULTIFUNC);

	if (s3_wait_fifo(7, 1))
	{
		
		return -E(ACCEL,NOSUP_NOW_MMAP);
	}
	
	outw(linp->x, ECR_CUR_X);
	outw(linp->y, ECR_CUR_Y);

	outw((max - 1), ECR_MAJAXIS_CNT);
	outw((min - max) << 1, ECR_DIAG_STEP);
	outw((min << 1), ECR_AXIAL_STEP);

	outw((min + min - max - (dx < 0)), ECR_ERROR_TERM);

	outw(EC07_LINE
	        | ((dy>0)            ? EC07_DIR(4) : 0)
		| ((dx>0)            ? EC07_DIR(1) : 0)
		| ((abs(dy)>abs(dx)) ? EC07_DIR(2) : 0)
		| EC07_DRAWPIXEL 
		| EC07_ACROSS_PLANE
		| 0x0001 ,ECR_COMMAND);

	s3_wait_idle(1);

	return(0);
}

/* KGI module interface */

int kgim_accel_init(struct kgi_card *card, struct kgi_display *dpy)
{
	NOTICE0("S3 accel driver loaded");
	/* !!! Check for Chipset and Version here !!! */
	



	return EOK;
}

void kgim_accel_done(struct kgi_display *dpy)
{
	NOTICE1("S3 accel driver removed");
}

int kgim_accel_check_mode (struct kgi_display *dpy, struct kgi_mode *mode,
			      enum kgi_tm_cmd *cmd)
{
	/* !!! We should probably check for accel-friendly dimensions here */
	return EOK;
}

void kgim_accel_set_mode (struct kgi_display *dpy, struct kgi_mode *mode)
{

	if (mode->request.graphtype != KGIGT_TEXT16) {

		/* Reset the Accelerator Engine */
		outw(((mode->request.graphtype == KGIGT_4BIT) 
		      ? (EC01_4BPP|EC01_ENH_ENB) : EC01_ENH_ENB) | EC01_LAW_ENB, ECR_ADVFUNC_CNTL);

		outw(EC00_RESET, ECR_SUBSYS_CNTL);
		inw_p(ECR_SUBSYS_STAT);	/* Wait a bit */
		outw(EC00_ENABLE, ECR_SUBSYS_CNTL);
		inw_p(ECR_SUBSYS_STAT); /* Wait a bit */

		outw(MF_MULT_MISC, ECR_MULTIFUNC); /* 16 bit I/O registers */

		/* Now set some basic accelerator registers */
		/* This might run in interrupt context, so maysched is 0 */
		s3_wait_fifo(6, 0);
		outw(SRC_FOREGROUND | MIX_SRC, ECR_FOREGR_MIX);
		outw(SRC_BACKGROUND | MIX_SRC, ECR_BACKGR_MIX);/* Direct color*/

		outw(MF_T_CLIP | 0, ECR_MULTIFUNC);	/* Clip virtual area  */
		outw(MF_L_CLIP | 0, ECR_MULTIFUNC);
		outw(MF_R_CLIP | (mode->request.virt.x - 1), ECR_MULTIFUNC);
		outw(MF_B_CLIP | (mode->request.virt.y - 1), ECR_MULTIFUNC);

		s3_wait_fifo(5, 0);
		outw(~0, ECR_WRITE_MASK);	/* Enable all planes */
		outw(~0, ECR_WRITE_MASK);	/* Enable all planes */
		outw(MF_PIX_CONTROL | MFA_SRC_FOREGR_MIX, ECR_MULTIFUNC);

		{
			struct kgi_box boxp;
			boxp.x = boxp.y = 0;
			boxp.w = mode->request.virt.x;
			boxp.h = mode->request.virt.y;
			s3_drawbox((struct kgi_graphic *) &dpy->dev, &boxp, 0);
		}
		/* !!! reset other registers too. */
	}
	return;
}

int kgim_accel_command (struct kgi_graphic *gr, kgi_uint cmd, kgi_uint arg)
{ 
	kgi_suggest *sugg;
	const char *name;

	if (!(gr->dev.state & KGI_MAPPED_ON_DISPLAY) ) {

		return -E(ACCEL,NOSUP_MODE_MMAP);
		/* Don't accel inactive screen! */
	}

	switch (cmd) { 
		
	case ACCEL_GETSUGGEST:
		sugg = (kgi_suggest *)(gr->io_buf);
		gr->io_res = gr->io_buf;

		if (sugg->id < 0 || sugg->id > 4) 
		{
			
			return -E(ACCEL,BOGUS);
		}
		
		name = sugg_default[gr->dev.mode.graphtype].text[sugg->id];
		if (!name)
		{
			
			return -E(ACCEL,BOGUS);
		}
		
		strcpy(sugg->name, name);
		sugg->args[0] = 0;
		sugg->id++;	/* Next is RAMDAC */
		if (sugg->id >= 4 || sugg_default[gr->dev.mode.graphtype].text[sugg->id] == NULL) 
		{
			
			sugg->id = 0;
		}
		
		return 0;

	case ACCEL_DRAWPIXEL:
#if 0	/* 1=SLOW */
		{  
			struct kgi_coord *h = gr->io_buf;
			unsigned offset = h->x + h->y * current_mode.virt.x;
			int col = 0x12345678;

			switch(current_mode.graphtype) {
			case KGIGT_8BIT:
			 	((char *)fb_graphics)[offset] = (char)col;
				return EOK;
			
			case KGIGT_15BIT:
			case KGIGT_16BIT:	
				((short *)fb_graphics)[offset] = (short)col;
				return EOK;

			case KGIGT_32BIT:
				((int *)fb_graphics)[offset] = (int)col;
				return EOK;
			default:
				printk("DUNNO MYSELF ... ???");
			}
		} 
		break;
#else
		return -E(ACCEL, NOSUP_ALWAYS_MMAP);
#endif

		/* Really accelerated functions. */

	case ACCEL_DRAWBOX:
		return(s3_drawbox(gr, gr->io_buf, 1));
		
	case ACCEL_COPYBOX:
		return(s3_copybox(gr, gr->io_buf));
		
	case ACCEL_DRAWLINE:
		return(s3_drawline(gr, gr->io_buf));

		/* Functions that will probably benefit 
		** from lower-lever accel 
		*/
		
	case ACCEL_FILLSCREEN:
		return -E(ACCEL,NOSUP_ALWAYS_LOWER);

		/* Functions that should probably use MMAP
		 */
	default: 
		return -E(ACCEL,NOSUP_ALWAYS_MMAP);

/* These functions cannot be supported via standard interfaces due to
 * pointers to userspace in the args, but are efficiently emulated via MMAP:
 * ACCEL_PUTHLINE	ACCEL_GETHLINE	ACCEL_PUTVLINE
 * ACCEL_GETVLINE	ACCEL_GETBOX		ACCEL_PUTBOX
 */

/* These are not implemented due to efficient support via MMAP.
 * ACCEL_PUTPIXEL	ACCEL_GETPIXEL	ACCEL_DRAWPIXEL (dummy for testing above)
 * ACCEL_DRAWCIRCLE	ACCEL_DRAWHLINE	ACCEL_DRAWVLINE
 */
	}
	return -E(ACCEL, NOSUP_ALWAYS_CANT);	/* The driver can't do this */
}
